import pytest

from twbm.bookmarks import (
    Bookmarks,
    parse_tags,
    check_tags,
    match_all_tags,
    match_any_tag,
    clean_tags,
    match_exact_tags,
)


@pytest.mark.parametrize(
    (
        "fts_query",
        "tags_all",
        "tags_all_not",
        "tags_any",
        "tags_any_not",
        "tags_exact",
        "result",
    ),
    (
        ("", None, None, None, None, "aaa,bbb", 2),
        ("xxxxx", None, None, None, None, None, 1),
        ("xxxxx", "ccc,xxx,yyy", None, None, None, None, 1),
        ("xxxxx", "xxx,yyy", None, None, None, None, 1),
        ("xxxxx", "xxx,yyy", "not1,not2", "xxx", None, "ccc,xxx,yyy", 1),
        ("xxxxx", "wrong", None, None, None, None, 0),
    ),
)
class TestBookmarks:
    def test_search_tags_exact(
        self,
        fts_query,
        tags_all,
        tags_all_not,
        tags_any,
        tags_any_not,
        tags_exact,
        result,
    ):
        bms = Bookmarks(fts_query=fts_query).filter(
            tags_all=tags_all,
            tags_all_not=tags_all_not,
            tags_any=tags_any,
            tags_any_not=tags_any_not,
            tags_exact=tags_exact,
        )
        assert len(bms) == result


@pytest.mark.parametrize(
    ("tags", "result"),
    (
        (("tag1", "tag2"), ",tag1,tag2,"),
        (("tag2", "tag1"), ",tag1,tag2,"),
        ((), ",,"),
    ),
)
def test_parse_tags(tags, result):
    assert parse_tags(tags) == result


@pytest.mark.parametrize(
    ("tags", "result"),
    (
        (("a", "b"), ["a", "b"]),
        (("xxx", "yyy"), []),
        (
            ("xxx", "yyy", "zzz"),
            [
                "zzz",
            ],
        ),
        ((), []),
    ),
)
def test_check_tags(dal, tags, result):
    # res = match_all_tags(('a', 'b'), ('a', 'b', 'c', 'd'))
    unknown_tags = check_tags(tags)
    assert unknown_tags == result
    _ = None


@pytest.mark.parametrize(
    ("tags", "bm_tags", "result"),
    (
        (("a", "b"), ("a", "b", "c", "d"), True),
        (("a", "b"), ("b", "c", "d"), False),
        (("a", "b"), ("b", "a"), True),
        (("a", "b"), ("a",), False),
    ),
)
def test_match_all_tags(tags, bm_tags, result):
    # res = match_all_tags(('a', 'b'), ('a', 'b', 'c', 'd'))
    assert match_all_tags(tags, bm_tags) is result


@pytest.mark.parametrize(
    ("tags", "bm_tags", "result"),
    (
        (("a", "b"), ("a", "b", "c", "d"), True),
        (("a", "b", "x"), ("a", "b", "c", "d"), True),
        (("a", "b", "x"), ("a",), True),
        (("a", "b"), ("x", "y"), False),
    ),
)
def test_match_any_tag(tags, bm_tags, result):
    # res = match_all_tags(('a', 'b'), ('a', 'b', 'c', 'd'))
    assert match_any_tag(tags, bm_tags) is result


def test_match_all(dal):
    bms = dal.get_bookmarks(fts_query="")
    tags = ("aaa", "ccc")
    # filtered = [bm for bm in bms if 'web' in bm.tags.split(',')]
    filtered = [bm for bm in bms if match_all_tags(tags, bm.split_tags)]
    assert len(filtered) >= 1


def test_match_any(dal):
    bms = dal.get_bookmarks(fts_query="")
    tags = ("aaa", "xxx")
    # filtered = [bm for bm in bms if 'web' in bm.tags.split(',')]
    filtered = [bm for bm in bms if match_any_tag(tags, bm.split_tags)]
    assert len(filtered) >= 4


@pytest.mark.parametrize(
    ("tags", "result"),
    (
        (("a", "b"), ["a", "b"]),
        (("b", "a"), ["a", "b"]),
        (("a", ",b"), ["a", "b"]),
        (("a,", ",b"), ["a", "b"]),
        (("a,xxx", ",b"), ["a", "b", "xxx"]),
        ((), []),
        (("a,", ",b", "A"), ["a", "b"]),
        (("a,A", ",b", "A"), ["a", "b"]),
    ),
)
def test_clean_tags(tags, result):
    # res = match_all_tags(('a', 'b'), ('a', 'b', 'c', 'd'))
    assert clean_tags(tags) == result


@pytest.mark.parametrize(
    ("tags", "bm_tags", "result"),
    (
        (("a", "b"), ("a", "b"), True),
        ((), (), True),
        (("a", "b"), ("a",), False),
    ),
)
def test_match_exact_tags(tags, bm_tags, result):
    # res = match_all_tags(('a', 'b'), ('a', 'b', 'c', 'd'))
    assert match_exact_tags(tags, bm_tags) is result



import logging
import os
import sys
import webbrowser
from os import isatty
from typing import Sequence, List

# import for nuitka
# noinspection PyUnresolvedReferences
import sqlalchemy.sql.default_comparator
import typer

from twbm.bookmarks import Bookmarks, clean_tags, parse_tags, check_tags
from twbm.buku import edit_rec, BukuDb
from twbm.db.dal import Bookmark, DAL
from twbm.environment import config
from twbm.handle_uri import open_it

_log = logging.getLogger(__name__)
log_fmt = r"%(asctime)-15s %(levelname)s %(name)s %(funcName)s:%(lineno)d %(message)s"
datefmt = "%Y-%m-%d %H:%M:%S"
logging.basicConfig(format=log_fmt, level=config.log_level, datefmt=datefmt)
logging.getLogger("urllib3").setLevel(logging.ERROR)

HELP_DESC = """
Generic URI manager for the command line.

Features:
- manages general URIs in sqlite database
- know how to open HTTP URLs, Microsoft Office Files, Images, ...
- can execute URIs as shell commands via the protocol prefix: 'shell::'
   Example: shell::vim +/"## SqlAlchemy" $HOME/document.md
"""

app = typer.Typer(help=HELP_DESC)

fts_sql = """
-- name: fts
-- record_class: Bookmark
"""


def _update_tags(
        ids: Sequence[int],
        tags: Sequence[str] = None,
        tags_not: Sequence[str] = None,
        force: bool = False,
):
    bms = Bookmarks(fts_query="").bms
    if tags is None:
        tags = ()
    if tags_not is None:
        tags_not = ()

    with DAL(env_config=config) as dal:
        for id in ids:
            bm = bms[id - 1]
            if force:
                new_tags = set(tags)
            else:
                new_tags = (set(bm.split_tags) | set(tags)) - set(tags_not)
            new_tags = sorted(list(new_tags))
            new_tags = f",{','.join(new_tags)},"
            _log.debug(f"{new_tags=}")
            bm.tags = new_tags
            dal.update_bookmark(bm)

            show_bms((bm,))


def show_bms(bms: Sequence[Bookmark], err: bool = True, show_timestamp: bool = False):
    for i, bm in enumerate(bms):
        offset = len(str(i)) + 2

        bmid_formatted = typer.style(
            f"{bm.id}", fg=typer.colors.BRIGHT_BLACK, bold=False
        )
        bmtitle_formatted = typer.style(
            f"{i}. {bm.metadata}", fg=typer.colors.GREEN, bold=True
        )
        typer.echo(f"{bmtitle_formatted} [{bmid_formatted}]", err=err)

        typer.secho(f"{' ':>{offset}}{bm.URL}", fg=typer.colors.YELLOW, err=err)
        if bm.desc != "":
            typer.secho(f"{' ':>{offset}}{bm.desc}", fg=None, err=err)
        typer.secho(
            f"{' ':>{offset}}{', '.join((tag for tag in bm.split_tags if tag != ''))}",
            fg=typer.colors.BRIGHT_BLUE,
            err=err,
        )
        if show_timestamp:
            typer.secho(
                f"{' ':>{offset}}{bm.last_update_ts.strftime('%Y-%m-%d %H:%M:%S')}",
                fg=typer.colors.BRIGHT_BLACK,
                err=err,
            )
        typer.secho("", err=err)
        typer.secho("", err=err)


def process(bms: Sequence[Bookmark]):
    help_text = """
        <n1> <n2>:      opens selection in browser
        p <n1> <n2>:    print id-list of selection
        p:              print all ids
        d <n1> <n2>:    delete selection
        e:              edit selection
        q:              quit
        h:              help
    """
    typer.secho(f"Selection: ", fg=typer.colors.GREEN, err=True)
    selection = [x for x in input().split()]

    try:
        # open it if no command letter
        try:
            selection = [int(x) for x in selection]
            for i in selection:
                uri = bms[i].URL
                open_it(uri)
            return
        except ValueError as e:
            pass

        # with command letter
        cmd = str(selection[0])
        selection = sorted([int(x) for x in selection[1:]])
        ids = list()

        if cmd == "p":
            if len(selection) == 0:
                ids = [bm.id for bm in bms]  # print all ids to stdout for piping
            else:
                for i in selection:
                    ids.append(bms[i].id)
            typer.echo(",".join((str(x) for x in ids)), err=False)  # stdout for piping

        elif cmd == "d":
            if len(selection) == 0:
                typer.echo(f"-W- no selection. Do nothing.")
                raise typer.Exit()
            else:
                for i in reversed(selection):  # must be reversed because of compacting
                    bm = bms[i]
                    typer.echo(bm.id)
                    _ = BukuDb(dbfile=config.dbfile).delete_rec(
                        index=bm.id, delay_commit=False
                    )
                    typer.echo(f"-M- Deleted entry: {bm.metadata}: {bm.URL}")

        elif cmd == "e":
            if len(selection) == 0:
                # typer.echo(f"-W- no selection. Do nothing.")
                # raise typer.Exit()
                for bm in bms:
                    _ = BukuDb(dbfile=config.dbfile).edit_update_rec(
                        index=bm.id, immutable=1
                    )
            else:
                for i in selection:
                    typer.echo(bms[i].id)
                    _ = BukuDb(dbfile=config.dbfile).edit_update_rec(
                        index=bms[i].id, immutable=1
                    )

        elif cmd == "h":
            typer.echo(help_text, err=True)

        elif cmd == "q":
            raise typer.Exit()

        else:
            typer.secho(f"-E- Invalid command {cmd}\n", err=True)
            typer.echo(help_text, err=True)
    except IndexError as e:
        typer.secho(
            f"-E- Selection {selection} out of range.",
            err=True,
            fg=typer.colors.RED,
        )
        raise typer.Abort()


@app.command()
def search(
        # ctx: typer.Context,
        fts_query: str = typer.Argument("", help="FTS query"),
        tags_exact: str = typer.Option(
            None, "-e", "--exact", help="match exact, comma separated list"
        ),
        tags_all: str = typer.Option(
            None, "-t", "--tags", help="match all, comma separated list"
        ),
        tags_any: str = typer.Option(
            None, "-T", "--Tags", help="match any, comma separated list"
        ),
        tags_all_not: str = typer.Option(
            None, "-n", "--ntags", help="not match all, comma separated list"
        ),
        tags_any_not: str = typer.Option(
            None, "-N", "--Ntags", help="not match any, comma separated list"
        ),
        tags_prefix: str = typer.Option(
            None, "--prefix", help="tags to prefix the tags option"
        ),
        non_interactive: bool = typer.Option(False, "--np", help="no prompt"),
        order_desc: bool = typer.Option(False, "-o", help="order by age, descending."),
        order_asc: bool = typer.Option(False, "-O", help="order by age, ascending."),
        verbose: bool = typer.Option(False, "-v", "--verbose"),
):
    """
    Searches bookmark database with full text search capabilities (FTS)
    (see: https://www.sqlite.org/fts5.html)

    Title, URL and description are FTS indexed. Tags are not part of FTS.

    Tags must be specified as comma separated list without blanks.
    Correct FTS search syntax: https://www.sqlite.org/fts5.html chapter 3.

    Example:\n
        twbm search 'security "single-page"'\n
        twbm search '"https://securit" *'\n
        twbm search '^security'\n
        twbm search 'postgres OR sqlite'\n
        twbm search 'security NOT keycloak'\n
        twbm search -t tag1,tag2 -n notag1 <searchquery>\n
        twbm search -e tag1,tag2\n
        twbm search xxxxx | twbm update -t x (interactive selection)\n

    \nCommands in interactive mode:\n
        <n1> <n2>:      opens selection in browser\n
        p <n1> <n1>:    prints corresponding id-list of selection\n
        p:              prints all ids\n
        d <n1> <n1>:    delete selection\n

            p <n1> <n2>:    print id-list of selection
            p:              print all ids
            d <n1> <n2>:    delete selection
            h:              help
    """
    if tags_prefix is not None:
        if tags_all is None:
            tags_all = tags_prefix
        else:
            tags_all = f"{tags_all},{tags_prefix}"

    if verbose:
        typer.echo(f"{config.twbm_db_url=}, {tags_all=}", err=True)

    bms = Bookmarks(fts_query=fts_query).filter(
        tags_all, tags_all_not, tags_any, tags_any_not, tags_exact
    )

    # ordering of results
    if order_desc:
        bms = sorted(bms, key=lambda bm: bm.last_update_ts)
    elif order_asc:
        bms = list(reversed(sorted(bms, key=lambda bm: bm.last_update_ts)))
    else:
        bms = sorted(bms, key=lambda bm: bm.metadata.lower() if bm.metadata else "")

    show_bms(bms, show_timestamp=order_desc or order_asc)
    typer.echo(f"Found: {len(bms)}", err=True)

    if not non_interactive:
        process(bms)
    else:
        ids = [bm.id for bm in bms]  # print all ids to stdout for piping
        typer.echo(",".join((str(x) for x in ids)), err=False)  # stdout for piping


@app.command()
def delete(
        # ctx: typer.Context,
        id_: int = typer.Argument(..., help="id to delete"),
        verbose: bool = typer.Option(False, "-v", "--verbose"),
):
    if verbose:
        typer.echo(f"Using DB: {config.twbm_db_url}", err=True)

    # use buku because of DB compactdb
    with DAL(env_config=config) as dal:
        bm = dal.get_bookmark_by_id(id_=id_)
        show_bms((bm,))
    _ = BukuDb(dbfile=config.dbfile).delete_rec(index=id_, delay_commit=False)


@app.command()
def update(
        # ctx: typer.Context,
        ids: str = typer.Argument(None, help="list of ids, separated by comma, no blanks"),
        tags: str = typer.Option(None, "-t", "--tags", help="add tags to taglist"),
        tags_not: str = typer.Option(None, "-n", "--tags", help="remove tags from taglist"),
        force: bool = typer.Option(
            False, "-f", "--force", help="overwrite taglist with tags"
        ),
        verbose: bool = typer.Option(False, "-v", "--verbose"),
):
    """
    Updates bookmarks with tags, either removes tags, add tags or overwrites entire taglist.

    Gotcha: in order to allow for piped input, ids must be separated by comma with no blanks.

    Example for using piped input:

        twbm search xxxxx | twbm update -t <tag>
    """
    if verbose:
        typer.echo(f"Using DB: {config.twbm_db_url}", err=True)
    if tags is not None:
        tags = tags.lower().replace(" ", "").split(",")
    if tags_not is not None:
        tags_not = tags_not.lower().replace(" ", "").split(",")

    # Gotcha: running from IDE looks like pipe
    is_pipe = not isatty(sys.stdin.fileno())
    ids_: Sequence[int] = list()

    if is_pipe:
        ids = sys.stdin.readline()

    try:
        ids = [int(x.strip()) for x in ids.split(",")]
    except ValueError as e:
        typer.secho(f"-E- Wrong input format.", color=typer.colors.RED, err=True)
        raise typer.Abort()

    print(ids)
    _update_tags(ids, tags, tags_not, force=force)


@app.command()
def open(
        # ctx: typer.Context,
        ids: str = typer.Argument(None, help="list of ids, separated by comma, no blanks"),
        verbose: bool = typer.Option(False, "-v", "--verbose"),
):
    """
    Opens bookmarks

    Gotcha: in order to allow for piped input, ids must be separated by comma with no blanks.

    Example for using piped input:

        twbm search xxxxx | twbm open
    """
    if verbose:
        typer.echo(f"Using DB: {config.twbm_db_url}", err=True)

    # Gotcha: running from IDE looks like pipe
    is_pipe = not isatty(sys.stdin.fileno())
    ids_: Sequence[int] = list()

    if is_pipe:
        ids = sys.stdin.readline()

    try:
        ids = [int(x.strip()) for x in ids.split(",")]
    except ValueError as e:
        typer.secho(f"-E- Wrong input format.", color=typer.colors.RED, err=True)
        raise typer.Abort()

    print(ids)
    with DAL(env_config=config) as dal:
        for id_ in ids:
            bm = dal.get_bookmark_by_id(id_=id_)
            show_bms((bm,))
            # TODO: generalize opening URI
            if bm.URL.startswith("http"):
                webbrowser.open(bm.URL, new=2)
            else:
                typer.secho(
                    "-W- Only HTTP implemented for direct open.",
                    color=typer.colors.RED,
                    err=True,
                )


@app.command()
def add(
        # ctx: typer.Context,
        url_data: List[str] = typer.Argument(..., help="URL and tags"),
        title: str = typer.Option(None, "--title"),
        desc: str = typer.Option("", "-d", "--desc"),
        edit: bool = typer.Option(False, "-e", "--edit", help="open in editor"),
        verbose: bool = typer.Option(False, "-v", "--verbose"),
        nofetch: bool = typer.Option(
            False, "-f", "--nofetch", help="do not try to fetch metadata from web"
        ),
):
    """
    Adds booksmarks to database and FTS index.

    Provide URL (required) and tags (optional) as parameters.

    Example:

        twbm add https://www.google.com tag1, tag2 --title "<title>"
    """
    if verbose:
        typer.echo(f"Using DB: {config.twbm_db_url}", err=True)
    url = url_data[0].strip().strip(",")
    tags = clean_tags(url_data[1:])

    if len(unknown_tags := check_tags(tags)) > 0:
        typer.confirm(f"Create {unknown_tags=} ?", abort=True)

    tags_in = parse_tags(tags)

    if edit:
        editor = os.environ.get("EDITOR", None)
        if editor is None:
            typer.secho(
                f"Editor not set, please set environment variable EDITOR",
                color=typer.colors.RED,
                err=True,
            )
            raise typer.Exit()
        result = edit_rec(
            editor=editor, url=url, title_in=title, tags_in=tags_in, desc=""
        )
        if result is not None:
            url, title, tags_in, desc_in = result
        else:
            raise typer.Abort()

    id_ = BukuDb(dbfile=config.dbfile).add_rec(
        url=url,
        title_in=title,
        tags_in=tags_in,
        desc=desc,
        immutable=0,
        delay_commit=False,
        fetch=(not nofetch),
    )
    with DAL(env_config=config) as dal:
        bm = dal.get_bookmark_by_id(id_=id_)
        show_bms((bm,))


@app.command()
def show(
        # ctx: typer.Context,
        id_: int = typer.Argument(..., help="id to print"),
        verbose: bool = typer.Option(False, "-v", "--verbose"),
):
    if verbose:
        typer.echo(f"Using DB: {config.twbm_db_url}", err=True)

    # _ = BukuDb(dbfile=config.dbfile).print_rec(index=id_)
    with DAL(env_config=config) as dal:
        bm = dal.get_bookmark_by_id(id_=id_)
        show_bms((bm,))


@app.command()
def edit(
        # ctx: typer.Context,
        id_: int = typer.Argument(..., help="id to edit"),
        verbose: bool = typer.Option(False, "-v", "--verbose"),
        nofetch: bool = typer.Option(
            False, "-f", "--nofetch", help="do not try to fetch metadata from web"
        ),
):
    immutable = -1 if nofetch else 1
    if verbose:
        typer.echo(f"Using DB: {config.twbm_db_url}", err=True)
    _ = BukuDb(dbfile=config.dbfile).edit_update_rec(index=id_, immutable=immutable)


@app.command()
def tags(
        # ctx: typer.Context,
        tag: str = typer.Argument(
            None,
            help="tag for which related tags should be shown. No input: all tags are printed.",
        ),
        verbose: bool = typer.Option(False, "-v", "--verbose"),
):
    """
    No parameter: Show all tags

    With tag as parameter: Show related tags, i.e. tags which are used in combination with tag.
    """
    if verbose:
        typer.echo(f"Using DB: {config.twbm_db_url}", err=True)

    if tag is not None:
        tag = tag.strip(",").strip().lower()

    with DAL(env_config=config) as dal:
        if tag is None:
            tags = dal.get_all_tags()
        else:
            tags = dal.get_related_tags(tag=tag)
        for t in tags:
            typer.echo(f"{t[1]:>4}: {t[0]}", err=True)


if __name__ == "__main__":
    _log.debug(config)
    app()



import pytest

from twbm.twb import (
    _update_tags,
)


@pytest.mark.parametrize(
    ("ids", "tags", "tags_not", "force", "result"),
    (
        ((1,), ("x",), ("ccc",), False, ",x,xxx,yyy,"),
        ((1,), ("x",), None, True, ",x,"),
        ((1,), None, None, False, ",ccc,xxx,yyy,"),
        ((1,), None, ("ccc", "xxx", "yyy"), False, ",,"),
    ),
)
def test__update_tags(dal, ids, tags, tags_not, result, force):
    # _update_tags((0,), ('x',), ('ob',))
    _update_tags(ids, tags, tags_not, force=force)
    assert dal.get_bookmarks(fts_query="xxxxx")[0].tags == result


@pytest.mark.skip("Not implemented yet")
def test_show_bmw():
    pass


@pytest.mark.skip("Not implemented yet")
def test_show_bmw():
    pass



import logging
from datetime import datetime
from typing import Sequence, Optional

import aiosql
import sqlalchemy as sa
from pydantic import BaseModel
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine, Connection

# from twbm.environment import Environment

_log = logging.getLogger(__name__)
logging.getLogger("sqlalchemy.engine").setLevel(logging.DEBUG)
# logging.getLogger('sqlalchemy').setLevel(logging.DEBUG)

metadata = sa.MetaData()

sim_results_table = sa.Table(
    "bookmarks",
    metadata,
    sa.Column("id", sa.Integer, primary_key=True),
    sa.Column("URL", sa.String(), nullable=False, unique=True),
    sa.Column("metadata", sa.String(), default=""),
    sa.Column("tags", sa.String(), default=""),
    sa.Column("desc", sa.String(), default=""),
    sa.Column("flags", sa.Integer(), default=0),
    sa.Column(
        "last_update_ts", sa.DateTime(), server_default=sa.func.current_timestamp()
    ),
)


class Bookmark(BaseModel):
    id: int = None
    URL: str = ""
    metadata: Optional[str] = ""
    tags: str = ",,"
    desc: str = ""
    flags: int = 0
    last_update_ts: datetime = datetime.utcnow()

    @property
    def split_tags(self) -> Sequence[str]:
        return [tag for tag in self.tags.split(",") if tag != ""]


# noinspection PyPropertyAccess
class DAL:
    _sql_alchemy_db_engine: Engine
    _conn: Connection

    is_simulated_environment: bool

    def __init__(self, env_config: "Environment"):
        self.bm_db_url = env_config.twbm_db_url
        self.record_classes = {
            "Bookmark": Bookmark,
        }

        # aiosql setup
        # sql_files_path = Path(__file__).parent.absolute() / Path("sql")
        # self.aiosql_queries = aiosql.from_path(
        #     f"{sql_files_path}", "sqlite3", record_classes=record_classes
        # )
        # self.aiosql_queries = aiosql.from_path(f"{sql_files_path}", "sqlite3")

    def __enter__(self):
        self._sql_alchemy_db_engine: Engine = create_engine(self.bm_db_url)
        self._conn = self._sql_alchemy_db_engine.connect()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self._conn.close()
        self._sql_alchemy_db_engine.dispose()

    @property
    def conn(self):
        return self._conn

    # TODO: not working
    def delete_bookmark(self, id: int) -> int:
        query = """
            -- name: delete_bookmark<!
            delete from bookmarks where id = :id
            returning *;
        """
        queries = aiosql.from_str(query, "sqlite3")
        result = queries.delete_bookmark(self.conn.connection, id=id)
        self.conn.connection.commit()
        return result

    def insert_bookmark(self, bm: Bookmark) -> int:
        query = """
            -- name: insert_bookmark<!
            -- record_class: Bookmark
            insert into bookmarks (URL, metadata, tags, desc, flags)
            values (:URL, :metadata, :tags, :desc, :flags)
            returning *;
        """
        queries = aiosql.from_str(query, "sqlite3", record_classes=self.record_classes)
        result = queries.insert_bookmark(
            self.conn.connection,
            URL=bm.URL,
            metadata=bm.metadata,
            tags=bm.tags,
            desc=bm.desc,
            flags=bm.flags,
        )
        self.conn.connection.commit()
        return result

    def update_bookmark(self, bm: Bookmark) -> int:
        query = """
            -- name: update_bookmark<!
            update bookmarks
            set tags = :tags
            where id = :id
            returning *;
        """
        queries = aiosql.from_str(query, "sqlite3")
        result = queries.update_bookmark(self.conn.connection, id=bm.id, tags=bm.tags)
        self.conn.connection.commit()
        return result

    def get_bookmark_by_id(self, id_: int) -> Bookmark:
        # Example query
        # noinspection SqlResolve
        query = """
            -- name: get_bookmark_by_id^
            -- record_class: Bookmark
            select *
            from bookmarks
            where id = :id;
            """
        queries = aiosql.from_str(query, "sqlite3", record_classes=self.record_classes)
        sql_result = queries.get_bookmark_by_id(self.conn.connection, id=id_)
        if not sql_result:
            # noinspection PyRedundantParentheses
            return Bookmark()
        return sql_result

    def get_bookmarks(self, fts_query: str) -> Sequence[Bookmark]:
        # Example query
        # noinspection SqlResolve
        if fts_query != "":
            query = """
                -- name: get_bookmarks
                -- record_class: Bookmark
                select *
                from bookmarks_fts
                where bookmarks_fts match :fts_query
                order by rank;
            """
            queries = aiosql.from_str(
                query, "sqlite3", record_classes=self.record_classes
            )
            sql_result = queries.get_bookmarks(
                self.conn.connection, fts_query=fts_query
            )
        else:  # TODO: make normal query
            query = """
                -- name: get_bookmarks
                -- record_class: Bookmark
                select *
                from bookmarks_fts
                order by rank;
            """
            queries = aiosql.from_str(
                query, "sqlite3", record_classes=self.record_classes
            )
            sql_result = queries.get_bookmarks(
                self.conn.connection, fts_query=fts_query
            )

        if not sql_result:
            # noinspection PyRedundantParentheses
            return (Bookmark(),)
        return sql_result

    def get_related_tags(self, tag: str):
        tag_query = f"%,{tag},%"
        # noinspection SqlResolve
        query = """
            -- name: get_related_tags
            with RECURSIVE split(tags, rest) AS (
                SELECT '', tags || ','
                FROM bookmarks
                WHERE tags LIKE :tag_query
                -- WHERE tags LIKE '%,ccc,%'
                UNION ALL
                SELECT substr(rest, 0, instr(rest, ',')),
                       substr(rest, instr(rest, ',') + 1)
                FROM split
                WHERE rest <> '')
            SELECT tags, count(tags) as n
            FROM split
            WHERE tags <> ''
            group by tags
            ORDER BY 2 desc;
        """
        queries = aiosql.from_str(query, "sqlite3")
        sql_result = queries.get_related_tags(self.conn.connection, tag_query=tag_query)

        # if not sql_result:
        #     # noinspection PyRedundantParentheses
        #     return (Bookmark(),)
        return sql_result

    def get_all_tags(self, with_frequency: bool = False):
        # noinspection SqlResolve
        query = """
            -- name: get_all_tags
            with RECURSIVE split(tags, rest) AS (
                SELECT '', tags || ','
                FROM bookmarks
                UNION ALL
                SELECT substr(rest, 0, instr(rest, ',')),
                       substr(rest, instr(rest, ',') + 1)
                FROM split
                WHERE rest <> '')
            SELECT tags, count(tags) as n
            FROM split
            WHERE tags <> ''
            group by tags
            ORDER BY 2 desc;
        """
        queries = aiosql.from_str(query, "sqlite3")
        sql_result = queries.get_all_tags(self.conn.connection)

        return sql_result
        # if with_frequency:
        #     return sql_result
        # else:
        #     return [tags[0] for tags in sql_result]



import logging
import os
import sqlite3
from os.path import exists
from tempfile import TemporaryDirectory

import pytest
import yaml

# from twbm.load_db import BukuDb
from twbm.buku import BukuDb

_log = logging.getLogger(__name__)
# log_fmt = r'%(asctime)-15s %(levelname)s %(name)s %(funcName)s:%(lineno)d %(message)s'
# logging.basicConfig(format=log_fmt, level=logging.DEBUG)

tests_data_path = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "tests_data"
)

# TEST_TEMP_DIR_OBJ = TemporaryDirectory(prefix='bukutest_')
TEST_TEMP_DIR_OBJ = TemporaryDirectory(prefix="bukutest_")
# TEST_TEMP_DIR_PATH = TEST_TEMP_DIR_OBJ.name
TEST_TEMP_DIR_PATH = "/tmp/bukutest"
TEST_TEMP_DBDIR_PATH = os.path.join(TEST_TEMP_DIR_PATH, "buku")
TEST_TEMP_DBFILE_PATH = os.path.join(TEST_TEMP_DBDIR_PATH, "bookmarks.db")

"""
test for sort order:
cls; python ./twbm/buku.py --db bm.db --np -t 'bb + py' --deep -- test
"""


@pytest.fixture()
def setup():
    if exists(TEST_TEMP_DBFILE_PATH):
        _log.info(f"Removing {TEST_TEMP_DBDIR_PATH}")
        os.remove(TEST_TEMP_DBFILE_PATH)


@pytest.fixture()
def db(request, setup):
    _log.info("setup")

    db = BukuDb(dbfile=TEST_TEMP_DBFILE_PATH)

    def teardown():
        _log.info("teardown")

    request.addfinalizer(teardown)

    return db


class PrettySafeLoader(
    yaml.SafeLoader
):  # pylint: disable=too-many-ancestors,too-few-public-methods
    def construct_python_tuple(self, node):
        return tuple(self.construct_sequence(node))


PrettySafeLoader.add_constructor(
    "tag:yaml.org,2002:python/tuple", PrettySafeLoader.construct_python_tuple
)


@pytest.fixture()
def chrome_db():
    # compatibility
    dir_path = os.path.dirname(os.path.realpath(__file__))
    res_yaml_file = os.path.join(dir_path, "tests_data", "25491522_res.yaml")
    res_nopt_yaml_file = os.path.join(dir_path, "tests_data", "25491522_res_nopt.yaml")
    json_file = os.path.join(dir_path, "tests_data", "Bookmarks")
    return json_file, res_yaml_file, res_nopt_yaml_file


def test_initdb(setup):
    assert not exists(TEST_TEMP_DBFILE_PATH)
    conn, curr = BukuDb.initdb(dbfile=TEST_TEMP_DBFILE_PATH)
    assert isinstance(conn, sqlite3.Connection)
    assert isinstance(curr, sqlite3.Cursor)
    assert exists(TEST_TEMP_DBFILE_PATH)
    curr.close()
    conn.close()


class ChromeTests:
    @pytest.mark.parametrize("add_pt", [True])
    def test_load_chrome_database(self, chrome_db, db, add_pt):
        """test method."""
        # compatibility
        json_file = chrome_db[0]

        db.load_chrome_database(json_file, None, add_pt)
        _ = None

    @pytest.mark.parametrize("add_pt", [True, False])
    def test_load_chrome_database_mock(self, chrome_db, db, add_pt):
        """test method."""
        # compatibility
        json_file = chrome_db[0]

        # with parent-tags and without
        res_yaml_file = chrome_db[1] if add_pt else chrome_db[2]

        with open(res_yaml_file, "r") as f:
            try:
                res_yaml = yaml.load(f, Loader=yaml.FullLoader)
            except RuntimeError:
                res_yaml = yaml.load(f, Loader=PrettySafeLoader)

        # this results in NOT having DB entries!!!
        # db.add_rec = mock.Mock()
        db.load_chrome_database(json_file, None, add_pt)
        # call_args_list_dict = dict(db.add_rec.call_args_list)
        # assert call_args_list_dict == res_yaml
        _ = None

    def test_traverse_bm_folder(self, db, data):
        bookmark_bar = data["roots"]["bookmark_bar"]
        for item in db.traverse_bm_folder(
            bookmark_bar["children"],
            unique_tag=None,
            folder_name="Bookmarks bar",
            add_parent_folder_as_tag=True,
        ):
            if "SearchPreview" in item[1]:
                assert item[2] == ",bookmarks bar,f+,xxx,yyy,"
            print(item)



import logging

import pytest

from twbm.handle_uri import open_it


@pytest.mark.skip("Only working when run in isolation, caplog problem?")
class TestOpenIt:
    @pytest.mark.parametrize(
        ("uri", "result"),
        (
            ("https://www.google.com", "https://www.google.com"),
            (
                "$HOME/dev/py/twbm/test/tests_data/test.pptx",
                "/Users/Q187392/dev/py/twbm/test/tests_data/test.pptx",
            ),
            ("/Users/Q187392/dev", "/Users/Q187392/dev"),
            ("~/dev", "/Users/Q187392/dev"),
            (
                "shell::vim +/'## SqlAlchemy' ./tests_data/sample_docu.md",
                "Running shell command",
            ),
        ),
    )
    def test_open_it(self, mocker, uri, result, caplog):
        # caplog.set_level(logging.DEBUG)
        with caplog.at_level(logging.DEBUG):
            webbrowser_open = mocker.patch("twbm.handle_uri.webbrowser.open")
            subprocess_run = mocker.patch("twbm.handle_uri.subprocess.run")
            open_it(uri)

            if uri.startswith("http"):
                webbrowser_open.assert_called()
            else:
                subprocess_run.assert_called()

        assert result in caplog.text


@pytest.mark.skip("Only runs in terminal interactively, see __main__")
def test_run_it():
    pass
